@@ -4,21 +4,28 @@ Module with several geodetic related methods
 """
 # import math
 # import sys
+import collections
+import typing
+
 import numpy as np
 
+import roktools.logger
+
 # WGS84 constants
 WGS84_A = 6378137.0
 WGS84_E = 8.1819190842622e-2
 
+DMS = collections.namedtuple('DMS', 'degrees minutes seconds')
+LLA = collections.namedtuple('LLA', 'longitude latitude altitude')
+XYZ = collections.namedtuple('XYZ', 'x y z')
+ElAz = collections.namedtuple('ElAz', 'elevation azimuth')
+
 class Coordinates(object):
 
-    def __init__(self, x, y, z):
+    def __init__(self, x:float, y:float, z:float) -> None:
+        self.point = XYZ(x, y, z)
 
-        self.x = x
-        self.y = y
-        self.z = z
-
-    def is_valid(self):
+    def is_valid(self) -> bool:
         """
         >>> coordinates = Coordinates(0, 0, 0)
         >>> coordinates.is_valid()
@@ -26,17 +33,47 @@ class Coordinates(object):
         """
         return valid_coordinates(xyz=self.xyz())
 
-    def xyz(self):
-        return [self.x, self.y, self.z]
+    def xyz(self) -> XYZ:
+        """
+        >>> coordinates = Coordinates(4889803.653, 170755.706, 4078049.851)
+        >>> coordinates.xyz()
+        XYZ(x=4889803.653, y=170755.706, z=4078049.851)
+        """
+        return self.point
 
-    def lla(self, radians=False, a=WGS84_A, e=WGS84_E):
-        return xyz_to_lla(self.x, self.y, self.z, radians=radians, a=a, e=e)
+    def lla(self, radians:bool=False, a:float=WGS84_A, e:float=WGS84_E) -> LLA:
+        """
+        >>> coordinates = Coordinates(4889803.653, 170755.706, 4078049.851)
+        >>> "lon: {:.0f} lat: {:.0f} height: {:.0f}".format(*coordinates.lla())
+        'lon: 2 lat: 40 height: 100'
+        """
+        return LLA(*xyz_to_lla(*self.point, radians=radians, a=a, e=e))
 
-    def __repr__(self):
-        return f'{self.x:.3f} {self.y:.3f} {self.z:.3f}'
+    def dms(self, radians:bool=False, a:float=WGS84_A, e:float=WGS84_E) -> LLA:
+        """
+        >>> coordinates = Coordinates(4889803.653, 170755.706, 4078049.851)
+        >>> coord = coordinates.dms()
+        >>> coord.longitude.degrees
+        1
+        >>> coord.longitude.minutes
+        59
+        """
+        lla = self.lla(radians=radians, a=a, e=e)
+
+        longitude = convert_decimal_units_to_dms(lla.longitude)
+        latitude = convert_decimal_units_to_dms(lla.latitude)
+
+        return LLA(longitude, latitude, lla.altitude)
+
+    def __repr__(self) -> str:
+        """
+        >>> Coordinates(4889803.653, 170755.706, 4078049.851)
+        4889803.653 170755.706 4078049.851
+        """
+        return f'{self.point.x:.3f} {self.point.y:.3f} {self.point.z:.3f}'
 
     @staticmethod
-    def from_lla(longitude_deg, latitude_deg, height_m, a=WGS84_A, e=WGS84_E):
+    def from_lla(longitude_deg:float, latitude_deg:float, height_m:float, a:float=WGS84_A, e:float=WGS84_E) -> 'Coordinates':
         """
         >>> coordinates = Coordinates.from_lla(2, 40, 100)
         >>> coordinates.is_valid()
@@ -48,7 +85,7 @@ class Coordinates(object):
         return Coordinates(*xyz)
 
     @staticmethod
-    def from_xyz(x, y, z):
+    def from_xyz(x:float, y:float, z:float) -> 'Coordinates':
         """
         >>> coordinates = Coordinates.from_lla(4889803.653, 170755.706, 4078049.851)
         >>> coordinates.is_valid()
@@ -56,7 +93,7 @@ class Coordinates(object):
         """
 
         return Coordinates(x, y, z)
-        
+
 
 def valid_coordinates(lla=None, xyz=None):
     """
@@ -127,24 +164,44 @@ def lla_to_xyz(longitude, latitude, height, a=WGS84_A, e=WGS84_E, radians=False)
     lon = longitude if radians else np.deg2rad(longitude)
     lat = latitude if radians else np.deg2rad(latitude)
 
-    sinlat = np.sin(lat);
-    coslat = np.cos(lat);
-    sinlon = np.sin(lon);
-    coslon = np.cos(lon);
-    e2 = e * e;
-
+    sinlat = np.sin(lat)
+    coslat = np.cos(lat)
+    sinlon = np.sin(lon)
+    coslon = np.cos(lon)
+    e2 = e * e
 
     # Intermediate calculation (prime vertical radius of curvature)
-    N = a / np.sqrt(1.0 - e2 * sinlat * sinlat);
-    nalt = N + height;
+    N = a / np.sqrt(1.0 - e2 * sinlat * sinlat)
+    nalt = N + height
 
-    x = nalt * coslat * coslon;
-    y = nalt * coslat * sinlon;
-    z = ((1.0-e2) * N + height) * sinlat;
+    x = nalt * coslat * coslon
+    y = nalt * coslat * sinlon
+    z = ((1.0-e2) * N + height) * sinlat
 
     return (x, y, z)
 
 
+def convert_decimal_units_to_dms(value:float) -> DMS:
+    """
+    Convert from longitude/latitude to Degrees, minutes and seconds
+    >>> latitude_deg = 41.528442316
+    >>> dms = convert_decimal_units_to_dms(latitude_deg)
+    >>> "{:.0f} {:.0f} {:.7f}".format(*dms)
+    '41 31 42.3923376'
+    >>> longitude_deg = 2.434318900
+    >>> dms = convert_decimal_units_to_dms(longitude_deg)
+    >>> "{:.0f} {:.0f} {:.7f}".format(*dms)
+    '2 26 3.5480400'
+    """
+
+    degrees = (int)(value)
+    minutes_float = (value - degrees) * 60
+    minutes = (int)(minutes_float)
+    seconds = (minutes_float - minutes) * 60
+
+    return DMS(degrees, minutes, seconds)
+
+
 #-------------------------------------------------------------------------------
 
 def xyz_to_lla(x, y, z, radians=False, a=WGS84_A, e=WGS84_E):
@@ -629,149 +686,147 @@ def ecef_to_enu(longitude, latitude, x, y, z, radians=False, matrix=None):
 #     return area
 # 
 # #-------------------------------------------------------------------------------
-# 
-# class Dop(object):
-#     """
-#     Class that implements computation of Dilution of Precision (DOP)
-# 
-# 
-#     Check that dop from NMEA GPGSV is the same as the one in NMEA GPGSA message
-#     (further investigations need to be conducted if NMEA generated by u-blox NEO 6P,
-#     discrepancies have been found)
-# 
-#     $GPGSA,A,3,10,07,05,02,29,04,08,13,,,,,1.72,1.03,1.38*0A
-#     $GPGSV,3,1,11,10,63,137,17,07,61,098,15,05,59,290,20,08,54,157,30*70
-#     $GPGSV,3,2,11,02,39,223,19,13,28,070,17,26,23,252,,04,14,186,14*79
-#     $GPGSV,3,3,11,29,09,301,24,16,09,020,,36,,,*76
-# 
-#     >>> elAz = np.array([[63,137],
-#     ...                  [61, 98],
-#     ...                  [59,290],
-#     ...                  [39,223],
-#     ...                  [ 9,301],
-#     ...                  [14,186],
-#     ...                  [54,157],
-#     ...                  [28, 70]])
-#     >>> dop_obj = Dop(elAz)
-#     >>> float('{0:.2f}'.format(dop_obj.pdop()))  # 1.72
-#     1.72
-#     >>> float('{0:.2f}'.format(dop_obj.vdop()))  # 1.38
-#     1.38
-#     >>> float('{0:.2f}'.format(dop_obj.hdop()))  # 1.03
-#     1.04
-# 
-# 
-#     Verification through the example of DOP computation proposed
-#     in Exercise 6-5 (page 228) of:
-# 
-#     Misra, P., Enge, P., "Global Positioning System: Signals,
-#     Measurements and Performance", 2nd Edition. 2006
-# 
-#     Increasing 30 degrees the azimuth position of 3 sats (+ one in zenith), doubles the vdop value
-#     leaving pdop approximately constant
-# 
-#     >>> elAz = np.array([[ 0,   0],
-#     ...                  [ 0, 120],
-#     ...                  [ 0, 240],
-#     ...                  [90,   0]])
-#     >>> dop_obj = Dop(elAz)
-#     >>> float('{0:.2f}'.format(dop_obj.vdop()))
-#     1.15
-#     >>> float('{0:.2f}'.format(dop_obj.hdop()))
-#     1.15
-# 
-#     >>> elAz = np.array([[30,   0],
-#     ...                  [30, 120],
-#     ...                  [30, 240],
-#     ...                  [90,  0]])
-#     >>> dop_obj = Dop(elAz)
-#     >>> float('{0:.2f}'.format(dop_obj.vdop()))
-#     2.31
-#     >>> float('{0:.2f}'.format(dop_obj.hdop()))
-#     1.33
-#     """
-# 
-#     def __init__(self, elAz, units='D'):
-# 
-#         self.elAz = elAz    # (numSats x 2) matrix containing elevation and azimuth measurements
-#         self.numSats = elAz.shape[0]
-#         self.units = units  # latitude, longitude, azimuth and elevation in degrees (D) or radians (R)
-#         self.__G()  # generate G matrix
-#         self.__Q()  # generate Q matrix
-# 
-# 
-#     def __G(self):
-#         """
-#         Construct geometry matrix G
-#         """
-# 
-#         rows = []
-#         for s in range(self.numSats):
-#             el = self.elAz[s, 0]
-#             az = self.elAz[s, 1]
-#             if self.units == 'D':
-#                 el = el * np.pi / 180
-#                 az = az * np.pi / 180
-# 
-#             gx = np.cos(el) * np.sin(az)
-#             gy = np.cos(el) * np.cos(az)
-#             gz = np.sin(el)
-# 
-#             rows.append([gx, gy, gz, 1])
-# 
-#         self.G = np.array(rows)
-# 
-# 
-#     def __Q(self):
-#         """
-#         Construct co-factor matrix
-#         """
-#         try:
-#             self.Q = np.linalg.inv(np.dot(np.matrix.transpose(self.G), self.G))
-#         except np.linalg.linalg.LinAlgError:
-#             self.Q = np.full([4, 4], np.nan)
-#             sys.stderr.write("except!\n")
-# 
-#     def gdop(self):
-#         """
-#         Geometric DOP
-#         """
-# 
-#         return np.sqrt(self.Q[0, 0] + self.Q[1, 1] + self.Q[2, 2] + self.Q[3, 3])
-# 
-# 
-#     def pdop(self):
-#         """
-#         Position DOP
-#         """
-# 
-#         return np.sqrt(self.Q[0, 0] + self.Q[1, 1] + self.Q[2, 2])
-# 
-# 
-#     def tdop(self):
-#         """
-#         Time DOP
-#         """
-# 
-#         return np.sqrt(self.Q[3, 3])
-# 
-# 
-#     def vdop(self):
-#         """
-#         Vertical DOP
-#         """
-# 
-#         return np.sqrt(self.Q[2, 2])
-# 
-# 
-#     def hdop(self):
-#         """
-#         Horizontal DOP
-#         """
-# 
-#         return np.sqrt(self.Q[0, 0] + self.Q[1, 1])
-# 
-# #-------------------------------------------------------------------------------
+
+
+class Dop(object):
+    """
+    Class that implements computation of Dilution of Precision (DOP)
+    
+
+    Check that dop from NMEA GPGSV is the same as the one in NMEA GPGSA message
+    (further investigations need to be conducted if NMEA generated by u-blox NEO 6P,
+    discrepancies have been found)
+
+    $GPGSA,A,3,10,07,05,02,29,04,08,13,,,,,1.72,1.03,1.38*0A
+    $GPGSV,3,1,11,10,63,137,17,07,61,098,15,05,59,290,20,08,54,157,30*70
+    $GPGSV,3,2,11,02,39,223,19,13,28,070,17,26,23,252,,04,14,186,14*79
+    $GPGSV,3,3,11,29,09,301,24,16,09,020,,36,,,*76
+
+    >>> elAz = [[63,137],
+    ...         [61, 98],
+    ...         [59,290],
+    ...         [39,223],
+    ...         [ 9,301],
+    ...         [14,186],
+    ...         [54,157],
+    ...         [28, 70]]
+    >>> dop = Dop(elAz)
+    >>> f'{dop.pdop():.2f}'
+    '1.72'
+    >>> f'{dop.vdop():.2f}'
+    '1.38'
+    >>> f'{dop.hdop():.2f}'
+    '1.04'
+
+
+    Verification through the example of DOP computation proposed
+    in Exercise 6-5 (page 228) of:
+
+    Misra, P., Enge, P., "Global Positioning System: Signals, 
+    Measurements and Performance", 2nd Edition. 2006
+
+    Increasing 30 degrees the azimuth position of 3 sats (+ one in zenith), doubles the vdop value
+    leaving pdop approximately constant
+
+    >>> el_az_list = [[ 0,   0],
+    ...               [ 0, 120],
+    ...               [ 0, 240],
+    ...               [90,   0]]
+    >>> dop = Dop(el_az_list)
+    >>> f'{dop.vdop():.2f}'
+    '1.15'
+    >>> f'{dop.hdop():.2f}'
+    '1.15'
+
+    >>> el_az_list = [[30,   0],
+    ...               [30, 120],
+    ...               [30, 240],
+    ...               [90,  0]]
+    >>> dop = Dop(el_az_list)
+    >>> f'{dop.vdop():.2f}'
+    '2.31'
+    >>> f'{dop.hdop():.2f}'
+    '1.33'
+    """
+    
+    def __init__(self, el_az_list:typing.List[ElAz], angles_in_degrees:bool=True) -> None:
+        
+        self.el_az_list = el_az_list   
+        self.angles_in_degrees = angles_in_degrees 
+        self.__G()  # generate G matrix
+        self.__Q()  # generate Q matrix
+
+
+    def __G(self) -> None:
+        """ 
+        Construct geometry matrix G
+        """
+
+        elevations = [v[0] for v in self.el_az_list]
+        azimuths = [v[1] for v in self.el_az_list]
+        if self.angles_in_degrees:
+            elevations = np.radians(elevations)
+            azimuths = np.radians(azimuths)
+
+        n_satellites = len(self.el_az_list)
+        
+        gx = np.cos(elevations) * np.sin(azimuths)
+        gy = np.cos(elevations) * np.cos(azimuths)
+        gz = np.sin(elevations)
+
+        rows = list(zip(gx, gy, gz, [1]*n_satellites))
+
+        self.G = np.array(rows)
+
+
+    def __Q(self) -> None:
+        """
+        Construct co-factor matrix
+        """
+        try:
+            self.Q = np.linalg.inv(np.dot(np.matrix.transpose(self.G), self.G))
+        except np.linalg.linalg.LinAlgError as e:   
+            self.Q = np.full([4, 4], np.nan)
+            roktools.logger.warning("Could not compute DOP. Reason {}".format(e))
+
+    def gdop(self) -> float:
+        """
+        Geometric DOP
+        """
+
+        return np.sqrt(self.Q[0, 0] + self.Q[1, 1] + self.Q[2, 2] + self.Q[3, 3])
+
+
+    def pdop(self) -> float:
+        """
+        Position DOP
+        """
+
+        return np.sqrt(self.Q[0, 0] + self.Q[1, 1] + self.Q[2, 2])
+
+
+    def tdop(self) -> float:
+        """
+        Time DOP
+        """
+
+        return np.sqrt(self.Q[3, 3])
+        
+
+    def vdop(self) -> float:
+        """
+        Vertical DOP
+        """
+
+        return np.sqrt(self.Q[2, 2])
+
+
+    def hdop(self) -> float:
+        """
+        Horizontal DOP
+        """
+
+        return np.sqrt(self.Q[0, 0] + self.Q[1, 1])
 
 
 if __name__ == "__main__":
